@@ -1,17 +1,21 @@
-import datetime
 import inspect
 import typing as t
+from functools import wraps
 
+from google.protobuf import json_format
+from google.protobuf.descriptor import Descriptor
 from google.protobuf.descriptor import FieldDescriptor
-from google.protobuf.json_format import _Printer  # type: ignore  # noqa
+from google.protobuf.descriptor_pool import DescriptorPool
 from google.protobuf.message import Message
-from google.protobuf.pyext._message import Descriptor  # type: ignore
 from google.protobuf.timestamp_pb2 import Timestamp
 from pyspark.sql import Column
 from pyspark.sql.functions import col
 from pyspark.sql.functions import udf
 from pyspark.sql.types import *
 
+from pbspark._timestamp import _from_datetime
+from pbspark._timestamp import _to_datetime
+
 # Built in types like these have special methods
 # for serialization via MessageToDict. Because the
 # MessageToDict function is an intermediate step to
@@ -23,7 +27,10 @@ _MESSAGETYPE_TO_SPARK_TYPE_MAP: t.Dict[str, t.Type[DataType]] = {
 _MESSAGETYPE_TO_SPARK_TYPE_KWARGS_MAP: t.Dict[str, t.Dict[str, t.Any]] = {}
 
 # Protobuf types map to these CPP Types. We map
-# them to Spark types for generating a spark schema
+# them to Spark types for generating a spark schema.
+# Note that bytes fields are specified by the `type` attribute in addition to
+# the `cpp_type` attribute so there is special handling in the `get_spark_schema`
+# method.
 _CPPTYPE_TO_SPARK_TYPE_MAP: t.Dict[int, t.Type[DataType]] = {
     FieldDescriptor.CPPTYPE_INT32: IntegerType,
     FieldDescriptor.CPPTYPE_INT64: LongType,
@@ -36,51 +43,70 @@ _CPPTYPE_TO_SPARK_TYPE_MAP: t.Dict[int, t.Type[DataType]] = {
     FieldDescriptor.CPPTYPE_STRING: StringType,
 }
 
-# region Timestamp code from WKT
-_EPOCH_DATETIME = datetime.datetime.utcfromtimestamp(0)
-_NANOS_PER_MICROSECOND = 1000
 
-
-def _round_toward_zero(value, divider):
-    result = value // divider
-    remainder = value % divider
-    if result < 0 < remainder:
-        return result + 1
-    else:
-        return result
-
-
-def _to_datetime(message: Timestamp) -> datetime.datetime:
-    """Serialize a Timestamp to a python datetime."""
-    return _EPOCH_DATETIME + datetime.timedelta(
-        seconds=message.seconds,
-        microseconds=_round_toward_zero(message.nanos, _NANOS_PER_MICROSECOND),
-    )
-
-
-# endregion
-
-
-class Printer(_Printer):
-    """Subclass the protobuf printer to override serialization"""
+# region serde overrides
+class _Printer(json_format._Printer):  # type: ignore
+    """Printer override to handle custom messages and byte fields."""
 
     def __init__(self, custom_serializers=None, **kwargs):
         self._custom_serializers = custom_serializers or {}
         super().__init__(**kwargs)
 
     def _MessageToJsonObject(self, message):
-        """Override serialization to prioritize custom serializers."""
         full_name = message.DESCRIPTOR.full_name
         if full_name in self._custom_serializers:
             return self._custom_serializers[full_name](message)
         return super()._MessageToJsonObject(message)
 
+    def _FieldToJsonObject(self, field, value):
+        if (
+            field.cpp_type == FieldDescriptor.CPPTYPE_STRING
+            and field.type == FieldDescriptor.TYPE_BYTES
+        ):
+            return value
+        return super()._FieldToJsonObject(field, value)
+
+
+class _Parser(json_format._Parser):  # type: ignore
+    """Parser override to handle custom messages."""
+
+    def __init__(self, custom_deserializers=None, **kwargs):
+        self._custom_deserializers = custom_deserializers or {}
+        super().__init__(**kwargs)
+
+    def ConvertMessage(self, value, message, path):
+        full_name = message.DESCRIPTOR.full_name
+        if full_name in self._custom_deserializers:
+            return self._custom_deserializers[full_name](value)
+        return super().ConvertMessage(value, message, path)
+
+
+# protobuf converts to/from b64 strings, but we prefer to stay as bytes.
+# we handle bytes parser by decorating to handle byte fields first
+def _handle_bytes(func):
+    @wraps(func)
+    def wrapper(value, field, require_str=False):
+        if (
+            field.cpp_type == FieldDescriptor.CPPTYPE_STRING
+            and field.type == FieldDescriptor.TYPE_BYTES
+        ):
+            return value
+        return func(value, field, require_str)
+
+    return wrapper
+
+
+json_format._ConvertScalarFieldValue = _handle_bytes(  # type: ignore[attr-defined]
+    json_format._ConvertScalarFieldValue  # type: ignore[attr-defined]
+)
+
+# endregion
+
 
 class MessageConverter:
-    """Class for converting serialized protobuf messages into spark structs."""
-
     def __init__(self):
-        self._custom_serializers = {}
+        self._custom_serializers: t.Dict[str, t.Callable] = {}
+        self._custom_deserializers: t.Dict[str, t.Callable] = {}
         self._message_type_to_spark_type_map = _MESSAGETYPE_TO_SPARK_TYPE_MAP.copy()
         self._message_type_to_spark_type_kwargs_map = (
             _MESSAGETYPE_TO_SPARK_TYPE_KWARGS_MAP.copy()
@@ -118,16 +144,32 @@ class MessageConverter:
                 full_name
             ] = _MESSAGETYPE_TO_SPARK_TYPE_KWARGS_MAP[full_name]
 
+    def register_deserializer(self, message: t.Type[Message], deserializer: t.Callable):
+        full_name = message.DESCRIPTOR.full_name
+        self._custom_deserializers[full_name] = deserializer
+
+    def unregister_deserializer(self, message: t.Type[Message]):
+        full_name = message.DESCRIPTOR.full_name
+        self._custom_deserializers.pop(full_name, None)
+
+    # region timestamp
     def register_timestamp_serializer(self):
-        """Serialize Timestamps to datetimes instead of strings."""
         self.register_serializer(Timestamp, _to_datetime, TimestampType)
 
     def unregister_timestamp_serializer(self):
         self.unregister_serializer(Timestamp)
 
+    def register_timestamp_deserializer(self):
+        self.register_deserializer(Timestamp, _from_datetime)
+
+    def unregister_timestamp_deserializer(self):
+        self.unregister_deserializer(Timestamp)
+
+    # endregion
+
     def message_to_dict(
         self,
-        message,
+        message: Message,
         including_default_value_fields=False,
         preserving_proto_field_name=False,
         use_integers_for_enums=False,
@@ -135,7 +177,7 @@ class MessageConverter:
         float_precision=None,
     ):
         """Custom MessageToDict using overridden printer."""
-        printer = Printer(
+        printer = _Printer(
             custom_serializers=self._custom_serializers,
             including_default_value_fields=including_default_value_fields,
             preserving_proto_field_name=preserving_proto_field_name,
@@ -143,7 +185,24 @@ class MessageConverter:
             descriptor_pool=descriptor_pool,
             float_precision=float_precision,
         )
-        return printer._MessageToJsonObject(message)
+        return printer._MessageToJsonObject(message=message)
+
+    def parse_dict(
+        self,
+        value: dict,
+        message: Message,
+        ignore_unknown_fields: bool = False,
+        descriptor_pool: t.Optional[DescriptorPool] = None,
+        max_recursion_depth: int = 100,
+    ):
+        """Custom ParseDict using overridden parser."""
+        parser = _Parser(
+            custom_deserializers=self._custom_deserializers,
+            ignore_unknown_fields=ignore_unknown_fields,
+            descriptor_pool=descriptor_pool,
+            max_recursion_depth=max_recursion_depth,
+        )
+        return parser.ConvertMessage(value=value, message=message, path=None)
 
     def get_spark_schema(
         self,
@@ -162,7 +221,7 @@ class MessageConverter:
         if inspect.isclass(descriptor) and issubclass(descriptor, Message):
             descriptor_ = descriptor.DESCRIPTOR
         else:
-            descriptor_ = descriptor
+            descriptor_ = descriptor  # type: ignore[assignment]
         for field in descriptor_.fields:
             spark_type: DataType
             if field.cpp_type == FieldDescriptor.CPPTYPE_MESSAGE:
@@ -177,6 +236,12 @@ class MessageConverter:
                     )
                 else:
                     spark_type = self.get_spark_schema(field.message_type)
+            # protobuf converts to/from b64 strings, but we prefer to stay as bytes
+            elif (
+                field.cpp_type == FieldDescriptor.CPPTYPE_STRING
+                and field.type == FieldDescriptor.TYPE_BYTES
+            ):
+                spark_type = ByteType()
             else:
                 spark_type = _CPPTYPE_TO_SPARK_TYPE_MAP[field.cpp_type]()
             if field.label == FieldDescriptor.LABEL_REPEATED:
@@ -237,3 +302,41 @@ class MessageConverter:
         column = col(data) if isinstance(data, str) else data
         protobuf_decoder_udf = self.get_decoder_udf(message_type, options)
         return protobuf_decoder_udf(column)
+
+    def get_encoder(
+        self, message_type: t.Type[Message], options: t.Optional[dict] = None
+    ) -> t.Callable:
+        kwargs = options or {}
+
+        def encoder(s: dict) -> bytes:
+            message = message_type()
+            self.parse_dict(s, message, **kwargs)
+            return message.SerializeToString()
+
+        return encoder
+
+    def get_encoder_udf(
+        self, message_type: t.Type[Message], options: t.Optional[dict] = None
+    ) -> t.Callable:
+        return udf(
+            self.get_encoder(message_type=message_type, options=options),
+            BinaryType(),
+        )
+
+    def to_protobuf(
+        self,
+        data: t.Union[Column, str],
+        message_type: t.Type[Message],
+        options: t.Optional[dict] = None,
+    ) -> Column:
+        """Serialize spark structs to protobuf messages.
+
+        Given a column and protobuf message type, serialize
+        protobuf messages also using our custom serializers.
+
+        The ``options`` arg should be a dictionary for the kwargs passed
+        our parse_dict (same args as protobuf's ParseDict).
+        """
+        column = col(data) if isinstance(data, str) else data
+        protobuf_encoder_udf = self.get_encoder_udf(message_type, options)
+        return protobuf_encoder_udf(column)
